package ru.amse.vorobiev.lce.view.impl;

import java.awt.Graphics;
import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

import ru.amse.vorobiev.lce.elements.IGate;
import ru.amse.vorobiev.lce.elements.IPlug;
import ru.amse.vorobiev.lce.elements.IWire;
import ru.amse.vorobiev.lce.ui.JEditor;
import ru.amse.vorobiev.lce.view.ICircuitView;
import ru.amse.vorobiev.lce.view.IGateView;
import ru.amse.vorobiev.lce.view.IPlugView;

public abstract class AbstractGateView implements IGateView {
	protected Point myUpperLeftCorner;
	protected int myWidth;
	protected int myHeight;
	protected final String myName;
	protected final IGate myGate;
	protected List<Point> myInputs;
	protected List<Point> myOutputs;
	protected final ICircuitView myCircuitView;
	
	public AbstractGateView(Point point, int width, int height, IGate gate, ICircuitView circuitView) {
		myUpperLeftCorner = point;
		myWidth = width;
		myHeight = height;
		myName = gate.getType().toString();
		myGate = gate;
		ensureHeight();
		myInputs = new ArrayList<Point>();
		myOutputs = new ArrayList<Point>();
		myCircuitView = circuitView;
		myCircuitView.addViewByElement(gate, this);
		
		int inputCount = myGate.getInputCount();
		if (inputCount != 0) {
			int inputStep = myHeight/(inputCount + 1);
			Point startPoint = new Point(myUpperLeftCorner.x, myUpperLeftCorner.y + inputStep); 
			for (int i = 0; i < inputCount; i++) {
				Point plugPoint = (Point) startPoint.clone();
				plugPoint.x -= JEditor.GATE_TO_PLUG_WIRE_LENGTH;
				new PlugView(plugPoint, myGate.getInput(i), myCircuitView);
				myInputs.add((Point) startPoint.clone());
				startPoint.y += inputStep;
			}
		}
				
		int outputCount = myGate.getOutputCount();
		if (outputCount != 0) {
			int outputStep = myHeight/(outputCount + 1);
			Point startPoint = new Point(myUpperLeftCorner.x + myWidth,
										myUpperLeftCorner.y + outputStep); 
			for (int i = 0; i < outputCount; i++) {
				Point plugPoint = (Point) startPoint.clone();
				plugPoint.x += JEditor.GATE_TO_PLUG_WIRE_LENGTH;
				new PlugView(plugPoint, myGate.getOutput(i), myCircuitView);
				myOutputs.add((Point) startPoint.clone());
				startPoint.y += outputStep;
			}
		}
	}
	
	public String getName() {
		return myName;
	}
	
	public List<Point> getInputs() {
		return myInputs;
	}

	public List<Point> getOutputs() {
		return myOutputs;
	}	
	
	public int getHeight() {
		return myHeight;
	}
	public Point getUpperLeftCorner() {
		return (Point) myUpperLeftCorner.clone();
	}
	public int getWidth() {
		return myWidth;
	}
	
	public void setHeight(int height) {
		myHeight = height;
	}
	
	public void setUpperLeftCorner(Point point) {
		calculatePlugPositions(point);
		myUpperLeftCorner = point;
	}
	
	public void ensureHeight() {
		int inputCount = myGate.getInputCount();
		int outputCount = myGate.getOutputCount();
		int count = (inputCount > outputCount) ? inputCount : outputCount; 
		int newHeght = (count + 1) * JEditor.MIN_SPACE_BEETWEEN_PLUGS + 5;
		if (myHeight != newHeght) {
			myHeight = (newHeght > JEditor.GATE_INITIAL_HEIGHT) 
						? newHeght
						: JEditor.GATE_INITIAL_HEIGHT;
		} 
	}
	
	public void calculatePlugPositions(Point point) {
		int dx = point.x - myUpperLeftCorner.x;
		int dy = point.y - myUpperLeftCorner.y;
		int inputCount = myGate.getInputCount();
		if (inputCount != 0) { 
			for (int i = 0; i < inputCount; i++) {
				Point input = myInputs.get(i);
				input.x += dx;
				input.y += dy;
				myInputs.set(i, input);
				IPlugView plugView = (IPlugView) myCircuitView.getViewByElement(myGate.getInput(i));
				Point center = plugView.getCenter();
				center.x += dx;
				center.y += dy;
				plugView.setCenter(center);
			}
		}
		
		int outputCount = myGate.getOutputCount();
		if (outputCount != 0) { 
			for (int i = 0; i < outputCount; i++) {
				Point output = myOutputs.get(i);
				output.x += dx;
				output.y += dy;
				myOutputs.set(i, output);
				IPlugView plugView = (IPlugView) myCircuitView.getViewByElement(myGate.getOutput(i));
				Point center = plugView.getCenter();
				center.x += dx;
				center.y += dy;
				plugView.setCenter(center);
			}
		}
	}
	
	private Point getNextInputPosition() {
		Point point = myInputs.get(myInputs.size() - 1);
		int inputStep = myHeight/(myInputs.size() + 1);
		if (inputStep < JEditor.MIN_SPACE_BEETWEEN_PLUGS) {
			inputStep = JEditor.MIN_SPACE_BEETWEEN_PLUGS;
		}
		return new Point(point.x - JEditor.GATE_TO_PLUG_WIRE_LENGTH, point.y + inputStep);
	}
	
	public void addInput() {
		IPlug plug = myGate.addInput();
		Point point = this.getNextInputPosition();
		myInputs.add(new Point(point.x + JEditor.GATE_TO_PLUG_WIRE_LENGTH, point.y));
		myCircuitView.addViewByElement(plug, new PlugView(point, plug, myCircuitView));
		this.ensureHeight();
		this.fixPlugPositions();
	}
	
	private Point getNextOutputPosition() {
		Point point = myOutputs.get(myOutputs.size() - 1);
		int outputStep = myHeight/(myOutputs.size() + 1);
		if (outputStep < JEditor.MIN_SPACE_BEETWEEN_PLUGS) {
			outputStep = JEditor.MIN_SPACE_BEETWEEN_PLUGS;
		}
		return new Point(point.x + JEditor.GATE_TO_PLUG_WIRE_LENGTH, point.y + outputStep);
	}
	
	public void addOutput() {
		IPlug plug = myGate.addOutput();
		Point point = this.getNextOutputPosition();
		myOutputs.add(new Point(point.x - JEditor.GATE_TO_PLUG_WIRE_LENGTH, point.y));
		myCircuitView.addViewByElement(plug, new PlugView(point, plug, myCircuitView));
		this.ensureHeight();
		this.fixPlugPositions();
	}
	
	public void deleteInput(int i) {
		IWire wire = myGate.getInput(i).getWire();
		if (wire != null) {
			IPlug output = wire.getFrom();
			output.setWire(null);
			myCircuitView.removeViewByElement(wire);
		}
		myGate.deleteInput(i);
		myInputs.remove(i);
		this.ensureHeight();
		this.fixPlugPositions();		
	}
	
	public void deleteOutput(int i) {
		IWire wire = myGate.getOutput(i).getWire();
		if (wire != null) {
			IPlug input = wire.getTo();
			input.setWire(null);
			myCircuitView.removeViewByElement(wire);
		}
		myGate.deleteOutput(i);
		myOutputs.remove(i);
		this.ensureHeight();
		this.fixPlugPositions();
	}
	
	private void fixPlugPositions() {
		int inputCount = myGate.getInputCount();
		if (inputCount != 0) {
			int inputStep = myHeight/(inputCount + 1);
			if (inputStep < JEditor.MIN_SPACE_BEETWEEN_PLUGS) {
				inputStep = JEditor.MIN_SPACE_BEETWEEN_PLUGS;
			}
			int y =  myUpperLeftCorner.y + inputStep; 
			for (int i = 0; i < inputCount; i++) {
				myInputs.get(i).y = y;
				((IPlugView) myCircuitView.getViewByElement(myGate.getInput(i))).getCenter().y = y;
				y += inputStep;
			}
		}
		int outputCount = myGate.getOutputCount();
		if (outputCount != 0) {
			int outputStep = myHeight/(outputCount + 1);
			if (outputStep < JEditor.MIN_SPACE_BEETWEEN_PLUGS) {
				outputStep = JEditor.MIN_SPACE_BEETWEEN_PLUGS;
			}
			int y =  myUpperLeftCorner.y + outputStep; 
			for (int i = 0; i < outputCount; i++) {
				myOutputs.get(i).y = y;
				((IPlugView) myCircuitView.getViewByElement(myGate.getOutput(i))).getCenter().y = y;
				y += outputStep;
			}
		}
	}
	
	public void setWidth(int width) {
		myWidth = width;
	}
	
	public IGate getGate() {
		return myGate;
	}
	
	public abstract void paint(Graphics g);
}
